/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <fbx/fbx.pmh>

#define FE_SFB_DEBUG	FALSE
#define FE_SFB_VERBOSE	FALSE

namespace fe
{

SurfaceAccessibleFbx::SurfaceAccessibleFbx(void)
{
}

SurfaceAccessibleFbx::~SurfaceAccessibleFbx(void)
{
}

SurfaceAccessibleFbx::FileContext::FileContext(void)
{
	//* The first thing to do is to create the FBX Manager which is
	//* the object allocator for almost all the classes in the SDK
	m_pFbxManager=FbxManager::Create();
	if(m_pFbxManager)
	{
		//* Create an IOSettings object.
		//* This object holds all import/export settings.
		FbxIOSettings* fbxIOSettings=
				FbxIOSettings::Create(m_pFbxManager,IOSROOT);
		m_pFbxManager->SetIOSettings(fbxIOSettings);

		//* Load plugins from the executable directory (optional)
		FbxString path=FbxGetApplicationDirectory();
		m_pFbxManager->LoadPluginsDirectory(path.Buffer());

		//* Create an FBX scene.
		//* This object holds most objects imported/exported from/to files.
		m_pFbxScene=FbxScene::Create(m_pFbxManager,"My Scene");
	}
}

SurfaceAccessibleFbx::FileContext::~FileContext(void)
{
	m_pFbxScene=NULL;

	//* Delete the FBX Manager.
	//* All the objects that have been allocated using the FBX Manager and that
	//* haven't been explicitly destroyed are also automatically destroyed.
	if(m_pFbxManager)
	{
		m_pFbxManager->Destroy();
		m_pFbxManager=NULL;
	}
}

BWORD SurfaceAccessibleFbx::load(String a_filename,sp<Catalog> a_spSettings)
{
#if FE_SFB_DEBUG
	feLog("SurfaceAccessibleFbx::load file \"%s\"\n",a_filename.c_str());
#endif

	if(a_spSettings.isValid())
	{
		m_frame=a_spSettings->catalog<Real>("frame");
		m_spFilter=a_spSettings->catalog< sp<Component> >("filter");

		const String options=a_spSettings->catalog<String>("options");

#if FE_SFB_DEBUG
		feLog("SurfaceAccessibleFbx::load options \"%s\"\n",options.c_str());
#endif

		if(!options.empty())
		{
			setOptions(options);
		}
	}
	else
	{
		m_frame=0.0;
		m_spFilter=NULL;
	}

	m_fps=24.0;

	if(m_optionMap.find("fps")!=m_optionMap.end())
	{
		m_fps=atof(m_optionMap["fps"].c_str());
	}

	if(m_filename!=a_filename)
	{
		m_spContext=NULL;
		m_filename=a_filename;
	}

	if(m_spContext.isValid() &&
			(!m_spContext->m_pFbxManager || !m_spContext->m_pFbxScene))
	{
		m_spContext=NULL;
	}

	if(m_spContext.isNull())
	{
		setContext(a_filename);

		if(m_spContext.isNull())
		{
#if FE_SFB_DEBUG
			feLog("SurfaceAccessibleFbx::load NULL FileContext\n");
#endif

			return FALSE;
		}
	}

	FbxScene* pFbxScene=m_spContext.isValid()? m_spContext->m_pFbxScene: NULL;

	if(pFbxScene)
	{
		startJoints();
		readTransformations(pFbxScene->GetRootNode());
		completeJoints();
	}

	return (pFbxScene != NULL);
};

BWORD SurfaceAccessibleFbx::setContext(String a_filename)
{
	sp<Catalog> spMasterCatalog=registry()->master()->catalog();
	const String key="SurfaceAccessibleFbx:"+a_filename;

	m_spContext=spMasterCatalog->catalog< hp<Component> >(key);

	if(m_spContext.isValid())
	{
		return TRUE;
	}

#if FE_SFB_DEBUG
	feLog("SurfaceAccessibleFbx::setContext"
			" creating new FileContext\n");
#endif

	m_spContext=new FileContext();
	spMasterCatalog->catalog< hp<Component> >(key)=m_spContext;

	if(m_spContext.isNull())
	{
		feLog("SurfaceAccessibleFbx::setContext"
				" unable to create FileContext\n");
		return FALSE;
	}

	FbxManager* pFbxManager=m_spContext->m_pFbxManager;
	if(!pFbxManager)
	{
		feLog("SurfaceAccessibleFbx::setContext"
				" unable to create FbxManager\n");
		m_spContext=NULL;
		return FALSE;
	}

#if FE_SFB_DEBUG
	feLog("SurfaceAccessibleFbx::setContext"
			" Autodesk FBX SDK version %s\n",
			pFbxManager->GetVersion());
#endif

	if(!m_spContext->m_pFbxScene)
	{
		feLog("SurfaceAccessibleFbx::setContext"
				" unable to create FbxScene\n");
		m_spContext=NULL;
		return FALSE;
	}

	if(!loadScene(a_filename.c_str()))
	{
		feLog("SurfaceAccessibleFbx::setContext"
				" failed to load \"%s\"\n",a_filename.c_str());

		m_spContext=NULL;
		return FALSE;
	}

	return TRUE;
}

BWORD SurfaceAccessibleFbx::loadScene(const char* a_pFilename)
{
#if FE_SFB_DEBUG
	feLog("SurfaceAccessibleFbx::loadScene\n");
#endif

	//* Get the file version number generate by the FBX SDK.
	I32 sdkMajor=0;
	I32 sdkMinor=0;
	I32 sdkRevision=0;
	FbxManager::GetFileFormatVersion(sdkMajor,sdkMinor,sdkRevision);

#if FE_SFB_DEBUG
	feLog("  FBX format %d.%d.%d for SDK\n",
			sdkMajor,sdkMinor,sdkRevision);
#endif

	//* Create an importer.
	FbxImporter* pFbxImporter=
			FbxImporter::Create(m_spContext->m_pFbxManager,"");

	//* Initialize the importer by providing a filename.
	const BWORD importStatus=pFbxImporter->Initialize(
			a_pFilename,-1,m_spContext->m_pFbxManager->GetIOSettings());

	I32 fileMajor=0;
	I32 fileMinor=0;
	I32 fileRevision=0;
	pFbxImporter->GetFileVersion(fileMajor,fileMinor,fileRevision);

#if FE_SFB_DEBUG
	feLog("  FBX format %d.%d.%d for \"%s\"\n",
			fileMajor,fileMinor,fileRevision,a_pFilename);
#endif

	if(!importStatus)
	{
		FbxString errorString=pFbxImporter->GetStatus().GetErrorString();
		feLog("SurfaceAccessibleFbx::loadScene"
				"FbxImporter::Initialize() failed: \"%s\"\n",
				errorString.Buffer());

		if (pFbxImporter->GetStatus().GetCode() ==
				FbxStatus::eInvalidFileVersion)
		{
			feLog("  invalid file version %d.%d.%d vs SDK version %d.%d.%d\n",
					fileMajor,fileMinor,fileRevision,
					sdkMajor,sdkMinor,sdkRevision);
		}

		return FALSE;
	}

#if FE_SFB_DEBUG
	if (pFbxImporter->IsFBX())
	{
		//* From this point, it is possible to access animation stack
		//* information without the expense of loading the entire file.

		const I32 animStackCount=pFbxImporter->GetAnimStackCount();

		feLog("  current animation stack: \"%s\"\n",
				pFbxImporter->GetActiveAnimStackName().Buffer());

		for(I32 animStackIndex=0;animStackIndex<animStackCount;animStackIndex++)
		{
			FbxTakeInfo* pFbxTakeInfo=pFbxImporter->GetTakeInfo(animStackIndex);

			feLog("    stack %d/%d \"%s\"\n",
					animStackIndex,animStackCount,pFbxTakeInfo->mName.Buffer());
			feLog("      description \"%s\"\n",
					pFbxTakeInfo->mDescription.Buffer());

			//* Change the value of the import name if the animation stack
			//* should be imported under a different name.
			//* Set the value of the import state to false if the animation
			//* stack should be not be imported. 
			feLog("      import \"%s\" state '%s'\n",
					pFbxTakeInfo->mImportName.Buffer(),
					pFbxTakeInfo->mSelect? "true": "false");

			FbxTimeSpan fbxTimeSpan=pFbxTakeInfo->mLocalTimeSpan;
			FbxTime fbxTimeStart=fbxTimeSpan.GetStart();
			FbxTime fbxTimeStop=fbxTimeSpan.GetStop();
			const I32 msStart=fbxTimeStart.GetMilliSeconds();
			const I32 msStop=fbxTimeStop.GetMilliSeconds();

			const Real framesPerMs=(0.001*m_fps);
			const Real frameStart=msStart*framesPerMs;
			const Real frameStop=msStop*framesPerMs;

			feLog("      span %dms to %dms (frame %.6G to %.6G)\n",
					msStart,msStop,frameStart,frameStop);
		}

		//* Set the import states. By default, the import states are always
		//* set to true. The code below shows how to change these states.
//      IOS_REF.SetBoolProp(IMP_FBX_MATERIAL,        true);
//      IOS_REF.SetBoolProp(IMP_FBX_TEXTURE,         true);
//      IOS_REF.SetBoolProp(IMP_FBX_LINK,            true);
//      IOS_REF.SetBoolProp(IMP_FBX_SHAPE,           true);
//      IOS_REF.SetBoolProp(IMP_FBX_GOBO,            true);
//      IOS_REF.SetBoolProp(IMP_FBX_ANIMATION,       true);
//      IOS_REF.SetBoolProp(IMP_FBX_GLOBAL_SETTINGS, true);
	}
#endif

	FbxScene* pFbxScene=m_spContext->m_pFbxScene;

	// Import the scene.
	BWORD success=pFbxImporter->Import(pFbxScene);

	if(!success &&
			pFbxImporter->GetStatus().GetCode()==FbxStatus::ePasswordError)
	{
		feLog("  FBX is password protected\n");
	}

#if FE_SFB_DEBUG
	const I32 poseCount=pFbxScene->GetPoseCount();
	feLog("  poseCount %d\n",poseCount);
	for(I32 poseIndex=0;poseIndex<poseCount;poseIndex++)
	{
		FbxPose* pFbxPose=pFbxScene->GetPose(poseIndex);

		const String poseName=pFbxPose? pFbxPose->GetName(): "<NULL>";
		feLog("    pose %d/%d \"%s\"\n",poseIndex,poseCount,poseName.c_str());
	}
#endif

	// Destroy the importer.
	pFbxImporter->Destroy();

	return success;
}

void SurfaceAccessibleFbx::readTransformations(FbxNode* a_pFbxNode)
{
	const String name=a_pFbxNode->GetName();

	FbxNode* pParentNode=a_pFbxNode->GetParent();
	const String parentName=pParentNode? pParentNode->GetName(): "";

#if FE_SFB_VERBOSE
	feLog("\nnode \"%s\" parent \"%s\"\n",name.c_str(),parentName.c_str());
#endif

	const Real msPerFrame=(1000.0/m_fps);

	FbxTime fbxTime=FBXSDK_TIME_INFINITE;
	fbxTime.SetMilliSeconds(m_frame*msPerFrame);

	FbxAMatrix matrixGlobal=a_pFbxNode->EvaluateGlobalTransform(fbxTime);
	FbxAMatrix matrixLocal=a_pFbxNode->EvaluateLocalTransform(fbxTime);

	SpatialTransform xformWorld;
	for(U32 m=0;m<4;m++)
	{
		const FbxVector4 row=matrixGlobal.GetRow(m);
		set(xformWorld.column(m),row[0],row[1],row[2]);
	}

#if FE_SFB_VERBOSE
	feLog("world:\n%s\n",c_print(xformWorld));
#endif

	SpatialTransform refWorld;
	setIdentity(refWorld);	//* TODO

	SpatialTransform xformRelative;
	for(U32 m=0;m<4;m++)
	{
		const FbxVector4 row=matrixLocal.GetRow(m);
		set(xformRelative.column(m),row[0],row[1],row[2]);
	}

#if FE_SFB_VERBOSE
	feLog("local:\n%s\n",c_print(xformRelative));
#endif

//	xformRelative.translation()*=0.03;	//* HACK

	const Color color(1.0,1.0,0.0);			//* TODO
	const Color localScale(1.0,1.0,1.0);	//* TODO

	setJoint(name,parentName,xformRelative,localScale,
			refWorld,color);

	const I32 childCount=a_pFbxNode->GetChildCount();
	for(I32 childIndex=0;childIndex<childCount;childIndex++)
	{
		readTransformations(a_pFbxNode->GetChild(childIndex));
	}
}

} // namespace
