/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <network/network.pmh>

namespace fe
{

ChunkSender::ChunkSender(void)
{
	m_operating = false;
	m_live = false;
}

ChunkSender::~ChunkSender(void)
{
	stop();
}

void ChunkSender::start(sp<ChunkSpool> spChunkSpool, sp<Socket> spSocket)
{
	m_spChunkSpool = spChunkSpool;
	m_chunkSenderTask.m_spChunkSpool = spChunkSpool;
	m_chunkSenderTask.m_spSocket = spSocket;
	m_chunkSenderTask.m_pChunkSender = this;
	m_live = true;
	m_pChunkSenderThread = new Thread(&m_chunkSenderTask);
	m_operating = true;
}

void ChunkSender::stop(void)
{
	if(m_operating)
	{
		m_spChunkSpool->interrupt();
		m_pChunkSenderThread->join();
		delete m_pChunkSenderThread;
		m_operating = false;
	}
}

bool ChunkSender::live(void)
{
	Mutex::Guard guard(m_liveMutex);
	return m_live;
}

void ChunkSender::setLive(bool setting)
{
	Mutex::Guard guard(m_liveMutex);
	m_live = setting;
}


ChunkSenderTask::ChunkSenderTask(void)
{
	m_pChunkSender = NULL;
}

void ChunkSenderTask::operate(void)
{
	try
	{
		m_pChunkSender->setLive(true);
		sp<Chunk> spChunk;
		while(true)
		{
			spChunk = m_spChunkSpool->wait();
#if FE_CODEGEN<=FE_DEBUG
			feLogGroup("network","chunk send size %u\n", spChunk->size());
#endif
			if(spChunk->size() == 0)
			{
				feX(e_writeFailed,
					"fe::ChunkSenderThread",
					"todo: decide on what to do with zero size chunks");
			}
			U32 nsize = htonl(spChunk->size());
			int n;
			n = m_spSocket->write((const void *)&nsize, sizeof(U32));
			if(n != sizeof(U32))
			{
				feX(e_writeFailed,
					"fe::ChunkSenderThread","connection failed in size write");
			}
			n = m_spSocket->write(spChunk->buffer(), spChunk->size());
			if(n != (IWORD)(spChunk->size()))
			{
				feX(e_writeFailed,
					"fe::ChunkSenderThread","connection failed in chunk write");
			}
		}
	}
	catch(Poisoned &p)
	{
		m_pChunkSender->setLive(false);

#if FE_CODEGEN<=FE_DEBUG
		feLogGroup("network","ChunkSenderTask poisoned\n");
#endif
	}
	catch(Exception &e)
	{
		m_pChunkSender->setLive(false);
		e.log();
		if(e.getResult() == e_writeFailed)
		{
			feLogGroup("network","thread dying due to write failure\n");
		}
		else
		{
			feLogGroup("network","thread dying due to unknown failure\n");
#if FE_CODEGEN<=FE_DEBUG
			exit(1);
#else
			throw e;
#endif
		}
	}
}

} /* namespace */


