/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include <operator/operator.pmh>

namespace fe
{

void SurfaceWalkOp::initialize(void)
{
	catalog<String>("icon")="FE_alpha";

	catalog<I32>("firstIndex")=0;
	catalog<I32>("firstIndex","high")=100;
	catalog<I32>("firstIndex","max")=1e6;
	catalog<String>("firstIndex","label")="First Index";

	catalog<I32>("secondIndex")=1;
	catalog<I32>("secondIndex","high")=100;
	catalog<I32>("secondIndex","max")=1e6;
	catalog<String>("secondIndex","label")="Second Index";

	catalog<Real>("maxTurn")=60.0;
	catalog<Real>("maxTurn","max")=180.0;
	catalog<String>("maxTurn","label")="Max Turn Angle";

	catalog< sp<Component> >("Input Surface");

	catalog< sp<Component> >("Reference Surface");
	 catalog<bool>("Reference Surface","optional")=TRUE;

	catalog< sp<Component> >("Output Curve");
	catalog<String>("Output Curve","IO")="output";
}

void SurfaceWalkOp::handle(Record& a_rSignal)
{
	const I32 firstIndex=catalog<I32>("firstIndex");
	const I32 secondIndex=catalog<I32>("secondIndex");
	const Real maxTurn=catalog<Real>("maxTurn");

	const Real minDot=cos(maxTurn*fe::degToRad);

	if(firstIndex==secondIndex)
	{
		return;
	}

	sp<SurfaceAccessibleI> spInputAccessible;
	if(!access(spInputAccessible,"Input Surface")) return;

	sp<SurfaceAccessibleI> spRefAccessible;
	if(!access(spRefAccessible,"Reference Surface",e_quiet))
	{
		spRefAccessible=spInputAccessible;
	}

	sp<SurfaceAccessibleI> spOutputAccessible;
	if(!accessOutput(spOutputAccessible,a_rSignal)) return;

	sp<DrawI> spDrawOutput;
	if(!accessDraw(spDrawOutput,a_rSignal)) return;

	sp<SurfaceAccessorI> spInputPosition;
	if(!access(spInputPosition,spInputAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spRefPosition;
	if(!access(spRefPosition,spRefAccessible,
			e_point,e_position)) return;

	sp<SurfaceAccessorI> spRefVertices;
	if(!access(spRefVertices,spRefAccessible,
			e_primitive,e_vertices)) return;

	const I32 pointCount=spInputPosition->count();
	const I32 refPointCount=spRefPosition->count();

	//* TODO cache point indexes of path while reference unchanged

	Array< std::set<I32> > pointNeighbors(refPointCount);

	const I32 refPrimitiveCount=spRefVertices->count();
	for(I32 primitiveIndex=0;primitiveIndex<refPrimitiveCount;primitiveIndex++)
	{
		if(interrupted())
		{
			break;
		}

		const I32 subCount=spRefVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		I32 lastPointIndex=
				spRefVertices->integer(primitiveIndex,subCount-1);
		for(I32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=
					spRefVertices->integer(primitiveIndex,subIndex);

			pointNeighbors[pointIndex].insert(lastPointIndex);
			pointNeighbors[lastPointIndex].insert(pointIndex);

			lastPointIndex=pointIndex;
		}
	}

	std::set<I32> pointHit;

	I32 vertCount=0;
	I32 maxCount=1024;
	SpatialVector* pCurve=new SpatialVector[maxCount];

	pCurve[vertCount++]=spInputPosition->spatialVector(firstIndex);

	SpatialVector lastRefPoint=spRefPosition->spatialVector(firstIndex);
	pointHit.insert(firstIndex);

	I32 pointIndex=secondIndex;
	while(vertCount<maxCount)
	{
		if(interrupted())
		{
			break;
		}

		pCurve[vertCount++]=spInputPosition->spatialVector(pointIndex);

		const SpatialVector refPoint=spRefPosition->spatialVector(pointIndex);
		pointHit.insert(pointIndex);

		const SpatialVector lastDir=unitSafe(refPoint-lastRefPoint);

		Real bestDot= -1.0;
		I32 bestIndex= -1;

		const std::set<I32>& rNeighbors=pointNeighbors[pointIndex];
		for(std::set<I32>::iterator it=rNeighbors.begin();
				it!=rNeighbors.end();it++)
		{
			const I32 otherIndex=*it;
			if(otherIndex>=pointCount || pointHit.find(otherIndex)!=pointHit.end())
			{
				continue;
			}

			const SpatialVector otherPoint=
					spRefPosition->spatialVector(otherIndex);

			const SpatialVector candidateDir=unitSafe(otherPoint-refPoint);
			const Real candidateDot=dot(lastDir,candidateDir);

			if(candidateDot<minDot)
			{
				continue;
			}

			if(bestDot<candidateDot)
			{
				bestDot=candidateDot;
				bestIndex=otherIndex;
			}
		}

		if(bestIndex<0)
		{
			break;
		}

		pointIndex=bestIndex;

		lastRefPoint=refPoint;
	}

	if(vertCount>1)
	{
		spDrawOutput->drawLines(pCurve,NULL,vertCount,
				DrawI::e_strip,false,NULL);
	}

	delete[] pCurve;

	String summary;
	summary.sPrintf("%d points",vertCount);
	catalog<String>("summary")=summary;
}

} // namespace
