if __name__ == '__main__':
	import forge
else:
	import ctypes
	import shlex

	class Context:
		def __init__(self, name = "Context"):
#			print "Context.init " + name
			self.objName = name
			self.verbose = False
			self.nameCount = {}

			terminalDL = ctypes.CDLL("libfexTerminalDL.so")
			self.fe_terminal_execute = terminalDL.fe_terminal_execute

		def setVerbose(self, verbose):
			self.verbose = verbose

		def execute(self, command):
			size = 1024
			result = ctypes.create_string_buffer(size)

			if self.verbose:
				print "\n[36m" + command + "[0m"

			self.fe_terminal_execute(size,result,command)

			if self.verbose:
				print "[34m" + result.value + "[0m"

			return result.value

		def name(self):
			return self.objName

		def setFrame(self, frame):
			command = "set frame " + str(frame)
			self.execute(command)

		def manage(self, library):
#			print "Context.manage " + library

			command = "manage " + library
			self.execute(command)

		def unmanage(self, library):
#			print "Context.unmanage " + library

			command = "unmanage " + library
			self.execute(command)

		def load(self, path, name = False, options = ""):
			name = name or path

#			print "Context.load " + path + " name " + name

			command = "load %s %s '%s'" % (name, path, options)
			result = self.execute(command)

			if result != "loaded":
				return None

			return Surface(self, name)

		def master(self):
#			print "Context.master "

			return Master(self)

		def create(self, operator, name = False):
			name = name or operator

#			print "Context.create " + operator + " name " + name

			command = "create " + operator + " " + name
			self.execute(command)

			return Operator(self, name)

	class Master:
		def __init__(self, context):
#			print "Master.init "
			self.context = context

		def __del__(self):
#			print "Master.del "
			pass

		def set(self, attribute, prop, value, typename = False):
			if not typename:
				if type(value) is float:
					typename = "real"
				elif type(value) is int:
					typename = "integer"
				elif type(value) is bool:
					typename = "boolean"
				elif type(value) is str:
					typename = "string"
					value = value.replace('"', '\\"')
					value = "\"%s\"" % value
				elif type(value) is list and len(value) == 3:
					typename = "vector3"
					value = "\"%G %G %G\"" % (value[0], value[1], value[2])
				else:
					print "Master.set incompatible type " + str(type(value))
					return

#			print "Master.set \"" + attribute + "\" " + typename + " " + str(value)

			command = "set MASTER \"%s\" \"%s\" %s %s" % (attribute, prop, typename, str(value))
			self.context.execute(command)

		def get(self, attribute, prop):
#			print "Master.get " + attribute

			command = "get MASTER \"%s\" \"%s\"" % (attribute, prop)
			return self.context.execute(command)

		def __setitem__(self, attribute, value):
			self.set(attribute, "value", value)

		def __getitem__(self, attribute):
			return self.get(attribute, "value")

	class Operator:
		def __init__(self, context, name = "Operator"):
#			print "Operator.init " + name
			self.context = context
			self.objName = name

		def __del__(self):
#			print "Operator.del " + self.objName

			command = "release " + self.objName
			self.context.execute(command)

		def name(self):
			return self.objName

		def operate(self, input0 = None, input1 = None, input2 = None, input3 = None, name = False):
			name = name or (self.objName + ".out")

#			print "Operator.operate " + name

			command = "signal " + self.objName + " " + name

			if input0:
#				print "  input0 " + input0.name()
				command += " " + input0.name()
			if input1:
#				print "  input1 " + input1.name()
				command += " " + input1.name()
			if input2:
#				print "  input2 " + input2.name()
				command += " " + input2.name()
			if input3:
#				print "  input3 " + input3.name()
				command += " " + input3.name()

			self.context.execute(command)

			return Surface(self.context, name)

		def set(self, attribute, prop, value, typename = False):
			if not typename:
				if type(value) is float:
					typename = "real"
				elif type(value) is int:
					typename = "integer"
				elif type(value) is bool:
					typename = "boolean"
				elif type(value) is str:
					typename = "string"
					value = value.replace('"', '\\"')
					value = "\"%s\"" % value
				elif type(value) is list and len(value) == 3:
					typename = "vector3"
					value = "\"%G %G %G\"" % (value[0], value[1], value[2])
				else:
					print "Operator.set incompatible type " + str(type(value))
					return

#			print "Operator.set \"" + attribute + "\" " + typename + " " + str(value)

			command = "set %s \"%s\" \"%s\" %s %s" % (self.objName, attribute, prop, typename, str(value))
			self.context.execute(command)

		def get(self, attribute, prop):
#			print "Operator.get " + attribute

			command = "get %s \"%s\" \"%s\"" % (self.objName, attribute, prop)
			return self.context.execute(command)

		def keys(self):

			command = "keys %s" % self.objName
			response = self.context.execute(command)

			result = {}

			entries = shlex.split(response)
			for entry in entries:
				key,keyType = entry.split("=")
				result[key] = keyType

			return result

		def __call__(self, input0 = None, input1 = None, input2 = None, input3 = None, name = False):
			return self.operate(input0, input1, input2, input3)

		def __setitem__(self, attribute, value):
			self.set(attribute, "value", value)

		def __getitem__(self, attribute):
			return self.get(attribute, "value")

	class Surface:
		def __init__(self, context, name = "Surface"):
#			print "Surface.init " + name
			self.context = context
			self.objName = name

			if not name in context.nameCount:
				context.nameCount[name] = 0

			context.nameCount[name]+=1

		def __del__(self):
#			print "Surface.del " + self.objName

			self.context.nameCount[self.objName]-=1

			if self.context.nameCount[self.objName] == 0:
				command = "release " + self.objName
				self.context.execute(command)

		def name(self):
			return self.objName

		def save(self, path, options = ""):
#			print "Surface.save " + path

			command = "save %s %s \"%s\"" % (self.objName, path, options)
			self.context.execute(command)

		def count(self, element):
#			print "Surface.count %s" % element

			command = "count %s %s" % (self.objName, element)
			return self.context.execute(command)

		def specs(self, element):
#			print "Surface.specs %s" % element

			command = "specs %s %s" % (self.objName, element)
			result = self.context.execute(command)

			attributeList = result.split(",")
			attributeMap = {}
			for attribute in attributeList:
				if attribute != "":
					entry = attribute.split()
					attributeMap[entry[1]] = entry[0]
			return attributeMap;

		def access(self, element, attribute):
#			print "Surface.access %s %s" % (element, attribute)

			command = "access %s %s %s" % (self.objName, element, attribute)
			name = self.context.execute(command)

			return Accessor(self.context, attribute, name)

		def clone(self, name):
#			print "Surface.clone to %s" % (name)

			command = "clone %s %s" % (self.objName, name)
			result = self.context.execute(command)

			if result != "cloned":
				return None

			return Surface(self.context, name)

	class Accessor:
		def __init__(self, context, attribute, name = "Accessor"):
#			print "Accessor.init " + name
			self.context = context
			self.isVertices = (attribute == "<vertices>")
			self.objName = name

			if not name in context.nameCount:
				context.nameCount[name] = 0

			context.nameCount[name]+=1

		def __del__(self):
#			print "Accessor.del " + self.objName

			self.context.nameCount[self.objName]-=1

			if self.context.nameCount[self.objName] == 0:
				command = "release " + self.objName
				self.context.execute(command)

		def name(self):
			return self.objName

		def __getitem__(self, index):
#			print "Accessor.__getitem__ " + str(index)

			command = "get %s %d" % (self.objName, index)
			result = self.context.execute(command)

			if self.isVertices:
				return result.split()

			return result

		def __setitem__(self, index, value):
#			print "Accessor.__setitem__ " + str(index) + " " + str(value)

			command = "set %s %d" % (self.objName, index)
			return self.context.execute(command)
