//
// Copyright 2016 Pixar
//
// Licensed under the Apache License, Version 2.0 (the "Apache License")
// with the following modification; you may not use this file except in
// compliance with the Apache License and the following modification to it:
// Section 6. Trademarks. is deleted and replaced with:
//
// 6. Trademarks. This License does not grant permission to use the trade
//	names, trademarks, service marks, or product names of the Licensor
//	and its affiliates, except as required to comply with Section 4(c) of
//	the License and to reproduce the content of the NOTICE file.
//
// You may obtain a copy of the Apache License at
//
//	 http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the Apache License with the above modification is
// distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied. See the Apache License for the specific
// language governing permissions and limitations under the Apache License.

// originally from pxr/imaging/lib/hd/unitHydraDelegate.h
// adapted for Free Electron

#ifndef __usd_HydraDelegate__
#define __usd_HydraDelegate__

#include "pxr/imaging/hd/sceneDelegate.h"
#include "pxr/imaging/hd/tokens.h"
#include "pxr/imaging/pxOsd/tokens.h"

#include "pxr/base/gf/vec3f.h"
#include "pxr/base/gf/vec3d.h"
#include "pxr/base/gf/vec4f.h"
#include "pxr/base/gf/vec4d.h"
#include "pxr/base/gf/matrix4f.h"
#include "pxr/base/gf/matrix4d.h"
#include "pxr/base/vt/array.h"
#include "pxr/base/tf/staticTokens.h"

#define TEST_TOKENS		(geometryAndGuides)

TF_DECLARE_PUBLIC_TOKENS(TestTokens,TEST_TOKENS);

/// \class HydraDelegate
/// A simple delegate class for unit test driver.
class HydraDelegate: public HdSceneDelegate
{
	public:
				HydraDelegate(HdRenderIndex *parentIndex,
					SdfPath const& delegateID);

		void	SetUseInstancePrimVars(bool v) { m_hasInstancePrimVars = v; }

		void	SetRefineLevel(int level);

		void	SetCamera(GfMatrix4d const &viewMatrix,
					GfMatrix4d const &projMatrix);
		void	SetCamera(SdfPath const &id, GfMatrix4d const &viewMatrix,
					GfMatrix4d const &projMatrix);
		void	AddCamera(SdfPath const &id);

		bool	HasLight(SdfPath const &id);
		void	AddLight(SdfPath const &id, GlfSimpleLight const &light);
		void	SetLight(SdfPath const &id, TfToken const &key, VtValue value);

		void	AddRenderTask(SdfPath const &id);
		void	AddRenderSetupTask(SdfPath const &id);
		void	AddSimpleLightTask(SdfPath const &id);
		void	AddShadowTask(SdfPath const &id);
		void	AddSelectionTask(SdfPath const &id);

		void	AdjustRenderSetupTask(SdfPath const &id,BWORD a_lighting,
					const GfVec4f& viewport);

		void	SetTaskParam(SdfPath const &id,
					TfToken const &name, VtValue val);
		VtValue GetTaskParam(SdfPath const &id, TfToken const &name);

		HdTaskSharedPtrVector GetRenderTasks(void);

		enum Interpolation { VERTEX, UNIFORM, CONSTANT, FACEVARYING, VARYING };

		void	AddMesh(SdfPath const &id,
					GfMatrix4f const &transform,
					VtVec3fArray const &points,
					VtIntArray const &numVerts,
					VtIntArray const &verts,
					VtVec3fArray const &normals,
					PxOsdSubdivTags const &subdivTags,
					VtValue const &color,
					Interpolation colorInterpolation,
					bool guide=false,
					SdfPath const &instancerId=SdfPath(),
					TfToken const &scheme=PxOsdOpenSubdivTokens->catmark,
					TfToken const &orientation=HdTokens->rightHanded,
					bool doubleSided=false);

		void	AddBasisCurves(SdfPath const &id,
					VtVec3fArray const &points,
					VtIntArray const &curveVertexCounts,
					VtVec3fArray const &normals,
					TfToken const &basis,
					VtValue const &color,
					Interpolation colorInterpolation,
					VtValue const &width,
					Interpolation widthInterpolation,
					SdfPath const &instancerId=SdfPath());

		void	AddPoints(SdfPath const &id,
					VtVec3fArray const &points,
					VtVec3fArray const &normals,
					VtValue const &color,
					Interpolation colorInterpolation,
					VtValue const &width,
					Interpolation widthInterpolation,
					SdfPath const &instancerId=SdfPath());

		void	AddInstancer(SdfPath const &id,
					SdfPath const &parentId=SdfPath(),
					GfMatrix4f const &rootTransform=GfMatrix4f(1));

		void	SetInstancerProperties(SdfPath const &id,
					VtIntArray const &prototypeIndex,
					VtVec3fArray const &scale,
					VtVec4fArray const &rotate,
					VtVec3fArray const &translate);

		void	AddSurfaceShader(SdfPath const &id,
					std::string const &source,
					HdShaderParamVector const &params);

		void	AddTexture(SdfPath const& id,
					GlfTextureRefPtr const& texture);

		void	Remove(SdfPath const &id);

				/// Remove all
		void	Clear(void);

				/// Hides an rprim, invalidating all collections it was in.
		void	HideRprim(SdfPath const &id);

				/// Un-hides an rprim, invalidating all collections it was in.
		void	UnhideRprim(SdfPath const &id);

				/// set per-prim repr
		void	SetReprName(SdfPath const &id, TfToken const &reprName);

				/// set per-prim refine level
		void	SetRefineLevel(SdfPath const &id, int refineLevel);

				/// Marks an rprim in the RenderIndex as dirty
				/// with the given dirty flags.
		void	MarkRprimDirty(SdfPath path,
					HdChangeTracker::RprimDirtyBits flag);

		void	UpdatePositions(SdfPath const &id, float time);
		void	UpdateRprims(float time);
		void	UpdateInstancerPrimVars(float time);
		void	UpdateInstancerPrototypes(float time);

		void	UpdateCamera(SdfPath const &id,
					TfToken const &key, VtValue value);

		void	BindSurfaceShader(SdfPath const &rprimId,
					SdfPath const &shaderId)
				{ m_surfaceShaderBindings[rprimId] = shaderId; }

		GfRange3d GetExtents(void);

virtual	bool			IsInCollection(SdfPath const& id,
							TfToken const& collectionName);
virtual	HdMeshTopology	GetMeshTopology(SdfPath const& id);
virtual	HdBasisCurvesTopology GetBasisCurvesTopology(SdfPath const& id);
virtual	PxOsdSubdivTags	GetSubdivTags(SdfPath const& id);
virtual	GfRange3d		GetExtent(SdfPath const & id);
virtual	GfMatrix4d		GetTransform(SdfPath const & id);
virtual	bool			GetVisible(SdfPath const & id);
virtual	bool			GetDoubleSided(SdfPath const & id);
virtual	int				GetRefineLevel(SdfPath const & id);
virtual	VtValue			Get(SdfPath const& id, TfToken const& key);
virtual	TfToken			GetReprName(SdfPath const &id);
virtual	TfTokenVector	GetPrimVarVertexNames(SdfPath const& id);
virtual	TfTokenVector	GetPrimVarVaryingNames(SdfPath const& id);
virtual	TfTokenVector	GetPrimVarFacevaryingNames(SdfPath const& id);
virtual	TfTokenVector	GetPrimVarUniformNames(SdfPath const& id);
virtual	TfTokenVector	GetPrimVarConstantNames(SdfPath const& id);
virtual	TfTokenVector	GetPrimVarInstanceNames(SdfPath const& id);
virtual	int				GetPrimVarDataType(SdfPath const& id,
							TfToken const& key);
virtual	int				GetPrimVarComponents(SdfPath const& id,
							TfToken const& key);
virtual	VtIntArray		GetInstanceIndices(SdfPath const& instancerId,
							SdfPath const& prototypeId);
virtual	GfMatrix4d		GetInstancerTransform(SdfPath const& instancerId,
							SdfPath const& prototypeId);
virtual	std::string		GetSurfaceShaderSource(SdfPath const &shaderId);
virtual	TfTokenVector	GetSurfaceShaderParamNames(SdfPath const &shaderId);
virtual	HdShaderParamVector GetSurfaceShaderParams(SdfPath const &shaderId);
virtual	VtValue			GetSurfaceShaderParamValue(SdfPath const &shaderId,
							TfToken const &paramName);
virtual	HdTextureResource::ID GetTextureResourceID(SdfPath const& textureId);
virtual	HdTextureResourceSharedPtr GetTextureResource(SdfPath const& textureId);

private:

	struct Mesh
	{
			Mesh(void) { }
			Mesh(TfToken const &scheme,
				TfToken const &orientation,
				GfMatrix4f const &transform,
				VtVec3fArray const &points,
				VtIntArray const &numVerts,
				VtIntArray const &verts,
				VtVec3fArray const &normals,
				PxOsdSubdivTags const &subdivTags,
				VtValue const &color,
				Interpolation colorInterpolation,
				bool guide,
				bool doubleSided):
				scheme(scheme),
				orientation(orientation),
				transform(transform),
				points(points),
				numVerts(numVerts),
				verts(verts),
				normals(normals),
				subdivTags(subdivTags),
				color(color),
				colorInterpolation(colorInterpolation),
				guide(guide),
				doubleSided(doubleSided)									{}

			TfToken			scheme;
			TfToken			orientation;
			GfMatrix4f		transform;
			VtVec3fArray	points;
			VtIntArray		numVerts;
			VtIntArray		verts;
			VtVec3fArray	normals;
			PxOsdSubdivTags	subdivTags;
			VtValue			color;
			Interpolation	colorInterpolation;
			bool			guide;
			bool			doubleSided;
			TfToken			reprName;
	};

	struct Curves
	{
			Curves(void)													{}

			Curves(
				VtIntArray const &curveVertexCounts,
				VtIntArray const &curveIndices,
				VtVec3fArray const &points,
				VtVec3fArray const &normals,
				TfToken const &basis,
				VtValue const &color,
				Interpolation colorInterpolation,
				VtValue const &width,
				Interpolation widthInterpolation) :
				curveVertexCounts(curveVertexCounts),
				curveIndices(curveIndices),
				points(points),
				normals(normals),
				basis(basis),
				color(color),
				colorInterpolation(colorInterpolation),
				width(width),
				widthInterpolation(widthInterpolation)						{}

			VtIntArray		curveVertexCounts;
			VtIntArray		curveIndices;
			VtVec3fArray	points;
			VtVec3fArray	normals;
			TfToken			basis;
			VtValue			color;
			Interpolation	colorInterpolation;
			VtValue			width;
			Interpolation	widthInterpolation;
			TfToken			reprName;
	};

	struct Points
	{
			Points(void)													{}

			Points(VtVec3fArray const &points,
				VtVec3fArray const &normals,
				VtValue const &color,
				Interpolation colorInterpolation,
				VtValue const &width,
				Interpolation widthInterpolation):
				points(points),
				normals(normals),
				color(color),
				colorInterpolation(colorInterpolation),
				width(width),
				widthInterpolation(widthInterpolation)						{}

			VtVec3fArray	points;
			VtVec3fArray	normals;
			VtValue			color;
			Interpolation	colorInterpolation;
			VtValue			width;
			Interpolation	widthInterpolation;
			TfToken			reprName;
	};

	struct Instancer
	{
			Instancer(void)													{}

			Instancer(VtVec3fArray const &scale,
				VtVec4fArray const &rotate,
				VtVec3fArray const &translate,
				GfMatrix4f const &rootTransform):
				scale(scale),
				rotate(rotate),
				translate(translate),
				rootTransform(rootTransform)								{}

			VtVec3fArray			scale;
			VtVec4fArray			rotate;
			VtVec3fArray			translate;
			VtIntArray				prototypeIndices;
			GfMatrix4f				rootTransform;
			std::vector<SdfPath>	prototypes;
	};

	struct SurfaceShader
	{
			SurfaceShader(void)												{}

			SurfaceShader(std::string const &src,
				HdShaderParamVector const &pms):
				source(src),
				params(pms)													{}

			std::string			source;
			HdShaderParamVector	params;
	};

	struct Texture
	{
			Texture(void)													{}

			Texture(GlfTextureRefPtr const &tex):
				texture(tex)												{}
			GlfTextureRefPtr	texture;
	};

		std::map<SdfPath,Mesh>				m_meshes;
		std::map<SdfPath,Curves>			m_curves;
		std::map<SdfPath,Points>			m_points;
		std::map<SdfPath,Instancer>			m_instancers;
		std::map<SdfPath,SurfaceShader>		m_surfaceShaders;
		std::map<SdfPath,SdfPath>			m_surfaceShaderBindings;
		std::map<SdfPath,Texture>			m_textures;
		TfHashSet<SdfPath,SdfPath::Hash>	m_hiddenRprims;

		SdfPath								m_cameraId;
		bool								m_runShadows;
		bool								m_hasInstancePrimVars;
		int									m_refineLevel;
		std::map<SdfPath, int>				m_refineLevels;

		typedef TfHashMap<TfToken, VtValue, TfToken::HashFunctor> ValueCache;
		typedef TfHashMap<SdfPath, ValueCache, SdfPath::Hash> ValueCacheMap;

		ValueCacheMap						m_valueCacheMap;
};

#endif  // __usd_HydraDelegate__
