/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include <platform/platform.pmh>

//#include <boost/format.hpp>

#define FE_LOG_GUARD_LOGGER	TRUE
#define FE_LOG_GUARD_COUT	!FE_LOG_GUARD_LOGGER

namespace fe
{

#if FE_SAFE_COUNTED_MUTEX
Log::Log(void):
		m_cnt(0)
	{
		if(ms_mutexCount++ == 0)
		{
			FEASSERT(!ms_pMutex);
			ms_pMutex=new RecursiveMutex();
		}
		FEASSERT(ms_pMutex);
	}

Log::~Log(void)
	{
		FEASSERT(ms_pMutex);
		if(--ms_mutexCount == 0)
		{
			delete ms_pMutex;
			ms_pMutex=NULL;
		}
	}
FE_DL_PUBLIC RecursiveMutex* Log::ms_pMutex=NULL;
FE_DL_PUBLIC I32 Log::ms_mutexCount=0;
#else
Log::Log(void): m_cnt(0) {}
Log::~Log(void) {}
FE_DL_PUBLIC RecursiveMutex Log::ms_mutex;
#endif

FE_DL_PUBLIC void intrusive_ptr_add_ref(Log *a_log)
{
	if (a_log)
	{
		a_log->acquire();
	}
}

FE_DL_PUBLIC void intrusive_ptr_release(Log *a_log)
{
	if (!a_log->release())
	{
		delete a_log;
	}
}

void StdoutLog::log(const std::string &a_message,
	std::map<std::string,std::string> &a_attributes)
{
#if FE_LOG_GUARD_COUT
	RecursiveMutex::Guard guard(mutex());
#endif

	std::cout << a_message.c_str();
	std::cout.flush();
}

FE_DL_PUBLIC String PrefixLog::ms_prefix;

void PrefixLog::log(const std::string &a_message,
	std::map<std::string,std::string> &a_attributes)
{
	const String replaced=
			String(a_message.c_str()).replace("\n(.)","\n"+ms_prefix+"\\1");

#if FE_LOG_GUARD_COUT
	RecursiveMutex::Guard guard(mutex());
#endif

	std::cout << ms_prefix.c_str();
	std::cout << replaced.c_str();
	std::cout.flush();
}

void AnnotatedLog::output(const char* source,int size,bool right)
{
	char buffer[1024];

	const int len=strlen(source);
	const int start=(len>size)? len-size: 0;
	strncpy(buffer,&source[start],len-start);
	buffer[len-start]=0;
	if(len>size)
	{
		buffer[0]='^';
	}
	if(right)
	{
		std::cout << std::setw(size) << std::right << buffer;
	}
	else
	{
		std::cout << std::setw(size) << std::left << buffer;
	}
}

void AnnotatedLog::log(const std::string &a_message,
	std::map<std::string,std::string> &a_attributes)
{
#if FE_LOG_GUARD_COUT
	RecursiveMutex::Guard guard(mutex());
#endif

	if(m_newline)
	{
		const char* filename=a_attributes["file"].c_str();
		const char* basename=strrchr(filename,'/');

//		output(a_attributes["date"].c_str(),11,false);
//		std::cout << " ";
		output(a_attributes["time"].c_str(),8,false);
		std::cout << " ";
		output(a_attributes["group"].c_str(),18,false);
		std::cout << " ";
		output(basename? basename+1: filename,20,true);
		std::cout << ":";
		output(a_attributes["line"].c_str(),4,false);
		std::cout << " ";
	}

	std::cout << a_message.c_str();
	std::cout.flush();

	const char* message=a_message.c_str();
	const int len=strlen(message);
	m_newline=(message[len-1]=='\n');
}

void GroupLog::log(const std::string &a_message,
	std::map<std::string,std::string> &a_attributes)
{
#if FE_LOG_GUARD_COUT
	RecursiveMutex::Guard guard(mutex());
#endif

	if(m_newline)
	{
#if FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64
		std::cout << "[34m" ;
#endif
		output(a_attributes["group"].c_str(),18,false);
#if FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64
		std::cout << "[0m" ;
#endif
		std::cout << " ";
	}

	std::cout << a_message.c_str();
	std::cout.flush();

	const char* message=a_message.c_str();
	const int len=strlen(message);
	m_newline=(message[len-1]=='\n');
}

void vsPrintf(std::string &target, const char* fmt, va_list ap,int& size)
{
	if(size<0)
	{
		feX("(debug)vsPrintf size<0 indicates unheeded discontinuation");
	}
	if(!size)
	{
		// start at 80 bytes, seems reasonable (1 tty line)
		size=80;
	}
	char *tmpbuf = new char[size];
	FEASSERT(NULL != tmpbuf);

	int n=FEVSNPRINTF((char *)tmpbuf, size, fmt, ap);

	if(n> -1 && n<size)
	{
		size= -1;
		target = tmpbuf;
		delete [] tmpbuf;
		return;
	}
	if(n> -1)
	{
		size=n+1;
	}
	else
	{
		size*=2;
	}

	delete [] tmpbuf;
	target="";
}
#if FALSE
void vsPrintf(std::string &target, const char *fmt, va_list ap)
{

	int sz = 80;
	char *tmp = new char[sz];
	FEASSERT(NULL != tmp);
	while(true)
	{
		int n=FEVSNPRINTF((char *)tmp, sz, fmt, ap);

		if(n> -1 && n<sz) break;
		if(n> -1) sz=n+1;
		else sz*=2;

		delete [] tmp;
		tmp = new char[sz];
		FEASSERT(NULL != tmp);

		// NOTE calling vsnprintf again without va_end,va_start is undefined
		feX("(debug)vsPrintf warning, expanding buffer\n");
	}
	target = tmp;
	delete [] tmp;
}
#endif

Logger::Logger(void)
{
#if FE_SAFE_COUNTED_MUTEX
	if(ms_mutexCount++ == 0)
	{
		FEASSERT(!ms_pMutex);
		ms_pMutex=new RecursiveMutex();
	}
	FEASSERT(ms_pMutex);
#endif

	String value;
	if(System::getEnvironmentVariable("FE_ANNOTATE", value) &&
			atoi(value.c_str()))
	{
		m_directLog = new AnnotatedLog();
	}
	else
	{
		m_directLog = new StdoutLog();
	}

	setLog("stdout", m_directLog);
	setLog("win32", new Win32Log());
	bind(".*", "stdout");
	bind(".*", "win32");
}

#if FE_SAFE_COUNTED_MUTEX
Logger::~Logger(void)
	{
		FEASSERT(ms_pMutex);
		if(--ms_mutexCount == 0)
		{
			delete ms_pMutex;
			ms_pMutex=NULL;
		}
	}
FE_DL_PUBLIC RecursiveMutex* Logger::ms_pMutex=NULL;
FE_DL_PUBLIC I32 Logger::ms_mutexCount=0;
#else
Logger::~Logger(void) {}
FE_DL_PUBLIC RecursiveMutex Logger::ms_mutex;
#endif

#if 0
void Logger::direct(const char *a_format, ...) const
{
	if(m_directLog.get()) { return; }

	std::string output;
	va_list ap;
	int size=0;
	while(size>=0)
	{
		va_start(ap, a_format);
		vsPrintf(output,a_format,ap,size);
		va_end(ap);
	}

	m_directLog->log(output);
}
#endif


void Logger::log(const String &a_group,
	std::map<std::string,std::string> &a_attributes, const char *a_format, ...)
{
#if FE_LOG_GUARD_LOGGER
	RecursiveMutex::Guard guard(mutex());
#endif

	std::map<String, std::vector<t_logptr> >::const_iterator
		i_bind = m_bindmap.find(a_group);
	if(i_bind == m_bindmap.end())
	{
		wiregroup(a_group);
		i_bind = m_bindmap.find(a_group);
		if(i_bind == m_bindmap.end())
		{
			return;
		}
	}

	std::string output;
	va_list ap;
	int size=0;
	while(size>=0)
	{
		va_start(ap, a_format);
		vsPrintf(output,a_format,ap,size);
		va_end(ap);
	}

	for(unsigned int i = 0; i < i_bind->second.size();i++)
	{
		i_bind->second[i]->log(output, a_attributes);
	}
	return;
}


void Logger::bind(const String &a_group, const String &a_logname)
{
	m_bindings[a_group].push_back(a_logname);
	m_bindmap[a_group].clear();

	for(std::map<String, std::vector<t_logptr> >::const_iterator
		i_bind = m_bindmap.begin(); i_bind != m_bindmap.end(); i_bind++)
	{
		wiregroup(i_bind->first);
	}
}

void Logger::clear(const String &a_group)
{
	m_bindings[a_group].clear();

	Regex regex(a_group.c_str());

	for(std::map<String, std::vector<t_logptr> >::iterator
		i_bind = m_bindmap.begin(); i_bind != m_bindmap.end(); )
	{
		std::string s(i_bind->first.c_str());

		if(regex.search(s.c_str()))
		{
			std::map<String, std::vector<t_logptr> >::iterator i_erase =
				i_bind;
			i_bind++;
			m_bindmap.erase(i_erase);
		}
		else
		{
			i_bind++;
		}
	}
}

void Logger::setLog(const String &a_logname, t_logptr a_log)
{
	m_logs[a_logname] = a_log;
}

void Logger::wiregroup(const String &a_group)
{
	std::set<String> logset;
	m_bindmap[a_group].clear();

	for(std::map<String, std::vector<String> >::const_iterator
		i_binding = m_bindings.begin();
		i_binding != m_bindings.end(); i_binding++)
	{
		const char* pattern=i_binding->first.c_str();
		if(!strncmp(pattern,".*",2) || Regex(pattern).search(a_group.c_str()))
		{
			for(unsigned int i = 0; i < i_binding->second.size(); i++)
			{
				std::map<String, t_logptr>::const_iterator i_log =
						m_logs.find(i_binding->second[i]);
				if(i_log == m_logs.end())
				{
					continue;
				}

				logset.insert(i_log->first);

				//m_bindmap[a_group].push_back(i_log->second);
			}
		}
	}

	for(std::set<String>::iterator iLS = logset.begin();
		iLS != logset.end(); iLS++)
	{
		m_bindmap[a_group].push_back(m_logs[*iLS]);
	}
}

} /* namespace */

FE_DL_PUBLIC void feLogDirect(const char *format, ...)
{
	std::string output;
	va_list ap;
	int size=0;
	while(size>=0)
	{
		va_start(ap, format);
		fe::vsPrintf(output,format,ap,size);
		va_end(ap);
	}

	fprintf(stdout, "%s", output.c_str());
}


//fe::Logger feLogger;

//fe::Logger& feLogger = fe::debugLogger;

